
#include "framebuffer.h"
#include "math.h"
#include <iostream>
#include "scene.h"
#include "M33.h"

#include <tiffio.h>

using namespace std;

FrameBuffer::FrameBuffer(int u0, int v0, int _w, int _h) : 
	Fl_Gl_Window(u0, v0, _w, _h, 0) {

	w = _w;
	h = _h;
	pix = new unsigned int[w*h];
	zb = new float[w*h];
	ishw = 0;
}

void FrameBuffer::ClearZB() {


	for (int i = 0; i < w*h; i++)
		zb[i] = 0.0f;

}

void FrameBuffer::draw() {

	if (!ishw) {
		glDrawPixels(w, h, GL_RGBA, GL_UNSIGNED_BYTE, pix);
	}
	else {
		scene->RenderHW(scene->ppc, scene->hwfb);
		// read HW framebuffer into pix using glReadPixels
		glReadPixels(0, 0, w, h, GL_RGBA, GL_UNSIGNED_BYTE, pix);
		//		glDrawPixels(w, h, GL_RGBA, GL_UNSIGNED_BYTE, pix);
	}

}

int FrameBuffer::handle(int event) {

	switch (event)
	{
	case FL_KEYBOARD: {
		KeyboardHandle();
		return 0;
	}
	case FL_MOVE: {
		int u = Fl::event_x();
		int v = Fl::event_y();
		if (u < 0 || u > w - 1 || v < 0 || v > h - 1)
			return 0;
		cerr << u << " " << v  << "          \r";
		if (this == scene->hwfb) {
			// print the values that you care about (color)
			V3 vcolor;
			vcolor.SetColor(Get(u, v));
			cerr << u << " " << v << "; " << vcolor << "          \r";
		}
		if (this == scene->fb && scene->visfb) {
			float currz = GetZ(u, v);
			if (currz != 0) {
				V3 pP(.5f + (float)u, .5f + (float)v, currz);
				V3 P;
				scene->visfb->CopyImage(scene->oldvisfb);
				scene->ppc->UnProject(pP, P);
				cerr << u << " " << v << " " << P << "          \r";
				scene->visfb->Draw3DPoint(P, V3(1.0f, 1.0f, 1.0f), 7, scene->visppc);
			}
			else
				scene->visfb->CopyImage(scene->oldvisfb);
			scene->visfb->redraw();
		}
		else if (this == scene->visfb) {
			cerr << u << " " << v << " " << scene->ifb->spix[(h-1-v)*w+u].slsN << "          \r";
		}
		return 0;
	}
	default:
		return 0;
	}
	return 0;
}

void FrameBuffer::KeyboardHandle() {

	if (this != scene->fb && this != scene->hwfb) {
		return;
	}
	int key = Fl::event_key();
	float rstep = 10.0f;
	float tstep = 10.0f;
	int needRender = 1;
	switch (key) {
	case FL_Left: {
		scene->ppc->Pan(rstep);
		break;
	}
	case FL_Right: {
		scene->ppc->Pan(-rstep);
		break;
	}
	case FL_Up: {
		scene->ppc->Tilt(rstep);
		break;
	}
	case FL_Down: {
		scene->ppc->Tilt(-rstep);
		break;
	}
	case 'o': {
		scene->ppc->Roll(rstep);
		break;
	}
	case 'p': {
		scene->ppc->Roll(-rstep);
		break;
	}
	case 'r': {
		scene->ppc->C = scene->ppc->C + scene->ppc->b*tstep;
		break;
	}
	case 'f': {
		scene->ppc->C = scene->ppc->C - scene->ppc->b*tstep;
		break;
	}
	case 'a': {
		scene->ppc->C = scene->ppc->C - scene->ppc->a*tstep;
		break;
	}
	case 'd': {
		scene->ppc->C = scene->ppc->C + scene->ppc->a*tstep;
		break;
	}
	case 'w': {
		scene->ppc->C = scene->ppc->C + scene->ppc->GetVD()*tstep;
		break;
	}
	case 's': {
		scene->ppc->C = scene->ppc->C - scene->ppc->GetVD()*tstep;
		break;
	}
	default:
		cerr << "INFO: do not understand keypress" << endl;
		return;
	}
	if (needRender) {
		if (this == scene->fb)
			scene->Render();
		else
			scene->hwfb->redraw();
	}

}

void FrameBuffer::SetBGR(unsigned int bgr) {

	for (int uv = 0; uv < w*h; uv++)
		pix[uv] = bgr;

}



// load a tiff image to pixel buffer
void FrameBuffer::LoadTiff(char* fname) {
	TIFF* in = TIFFOpen(fname, "r");
	if (in == NULL) {
		cerr << fname << " could not be opened" << endl;
		return;
	}

	int width, height;
	TIFFGetField(in, TIFFTAG_IMAGEWIDTH, &width);
	TIFFGetField(in, TIFFTAG_IMAGELENGTH, &height);
	if (w != width || h != height) {
		w = width;
		h = height;
		delete[] pix;
		pix = new unsigned int[w*h];
		if (zb) {
			delete[]zb;
			zb = new float[w*h];
		}
		size(w, h);
		glFlush();
		glFlush();
	}

	if (TIFFReadRGBAImage(in, w, h, pix, 0) == 0) {
		cerr << "failed to load " << fname << endl;
	}

	TIFFClose(in);
}

// save as tiff image
void FrameBuffer::SaveAsTiff(char *fname) {

	TIFF* out = TIFFOpen(fname, "w");

	if (out == NULL) {
		cerr << fname << " could not be opened" << endl;
		return;
	}

	TIFFSetField(out, TIFFTAG_IMAGEWIDTH, w);
	TIFFSetField(out, TIFFTAG_IMAGELENGTH, h);
	TIFFSetField(out, TIFFTAG_SAMPLESPERPIXEL, 4);
	TIFFSetField(out, TIFFTAG_BITSPERSAMPLE, 8);
	TIFFSetField(out, TIFFTAG_ORIENTATION, ORIENTATION_TOPLEFT);
	TIFFSetField(out, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG);
	TIFFSetField(out, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_RGB);

	for (uint32 row = 0; row < (unsigned int)h; row++) {
		TIFFWriteScanline(out, &pix[(h - row - 1) * w], row);
	}

	TIFFClose(out);
}



void FrameBuffer::Set(int u, int v, unsigned int col) {

	int uv = (h - 1 - v)*w + u;
	pix[uv] = col;

}

unsigned int FrameBuffer::GetGuarded(V3 pp) {

	if (pp[0] < 0.0f || pp[0] >= (float)w)
		return 0x00000000;
	if (pp[1] < 0.0f || pp[1] >= (float)h)
		return 0x00000000;

	return Get((int)pp[0], (int)pp[1]);

}


unsigned int FrameBuffer::Get(int u, int v) {

	int uv = (h - 1 - v)*w + u;
	return pix[uv];

}

float FrameBuffer::GetZ(int u, int v) {

	if (u < 0 || u > w - 1 || v < 0 || v > h - 1)
		return FLT_MAX;
	int uv = (h - 1 - v)*w + u;
	return zb[uv];

}

void FrameBuffer::SetZ(int u, int v, float z) {

	int uv = (h - 1 - v)*w + u;
	zb[uv] = z;

}


void FrameBuffer::SetAsChecker(unsigned int col0, unsigned int col1, int csize) {

	for (int v = 0; v < h; v++) {
		for (int u = 0; u < w; u++) {
			int cu = u / csize;
			int cv = v / csize;
			if ((cu + cv) % 2) {
				Set(u, v, col0);
			}
			else
				Set(u, v, col1);
		}
	}

}



void FrameBuffer::Draw2DPoint(V3 p, int psize, unsigned int col) {

	float uf = p[0];
	float vf = p[1];
	int u0 = (int)uf;
	int v0 = (int)vf;
	for (int v = v0 - psize / 2; v <= v0 + psize / 2; v++)
		for (int u = u0 - psize / 2; u <= u0 + psize / 2; u++) {
			if (u < 0 || u > w - 1 || v < 0 || v > h - 1)
				continue;
			float oldz = GetZ(u, v);
			if (oldz > p[2])
				continue;
			SetZ(u, v, p[2]);
			Set(u, v, col);
		}
}


void FrameBuffer::Draw3DPoint(V3 P, V3 colv, int pSize, PPC *ppc) {


	V3 pP;
	if (!ppc->Project(P, pP))
		return;
	Draw2DPoint(pP, pSize, colv.GetColor());

}


void FrameBuffer::Draw3DSegment(V3 V0, V3 c0, V3 V1, V3 c1, PPC *ppc) {

	V3 pV0;
	if (!ppc->Project(V0, pV0))
		return;
	if (pV0[0] < 0.0f || pV0[0] > (float)w || pV0[1] < 0.0f || pV0[1] > (float) h)
		return;
	V3 pV1;
	if (!ppc->Project(V1, pV1))
		return;
	if (pV1[0] < 0.0f || pV1[0] > (float)w || pV1[1] < 0.0f || pV1[1] > (float) h)
		return;

	int samplesN;
	float uspan = fabsf(pV0[0] - pV1[0]);
	float vspan = fabsf(pV0[1] - pV1[1]);
	samplesN = (uspan < vspan) ? (int)vspan : (int)uspan;
	samplesN += 2;

	V3 cp;
	V3 dp = (pV1 - pV0) / (float)(samplesN - 1);
	V3 cc;
	V3 dc = (c1 - c0) / (float)(samplesN - 1);
	int si;
	for (si = 0, cp = pV0, cc = c0; si < samplesN; si++, cp = cp + dp, cc = cc + dc) {
		Draw2DPoint(cp, 1, cc.GetColor());
	}

}

void FrameBuffer::VisualizeSamples(PPC *ppc, PPC *visppc, FrameBuffer *visfb, float zPlane) {

	for (int v = 0; v < h; v++) {
		for (int u = 0; u < w; u++) {
			V3 pP(.5f + (float)u, .5f + (float)v, GetZ(u, v));
			if (zPlane != 0.0f)
				pP[2] = zPlane;
			V3 P;
			if (ppc->UnProject(pP, P) && Get(u, v) != 0xFFFFFFFF) {
				V3 colv; colv.SetColor(Get(u, v));
				visfb->Draw3DPoint(P, colv, 1, visppc);
			}
		}
	}


}


void FrameBuffer::CopyImage(FrameBuffer *fb) {

	for (int uv = 0; uv < w*h; uv++) {
		pix[uv] = fb->pix[uv];
		zb[uv] = fb->zb[uv];
	}
}


void FrameBuffer::SetFromImage(FrameBuffer *fromfb, PPC *fromppc, PPC *thisppc) {

	for (int v = 0; v < h; v++) {
		for (int u = 0; u < w; u++) {
//u = w / 2;
//v = h / 2;
			V3 P = thisppc->GetPixelCenter(u, v);
			V3 p;
			if (!fromppc->Project(P, p))
				continue;
			unsigned int colDest = fromfb->GetGuarded(p);
			if (colDest == 0x00000000)
				continue;
			unsigned colSrc = Get(u, v);
			if (colSrc == 0xFF808080)
				Set(u, v, colDest);
			else {
				V3 colDestv; colDestv.SetColor(colDest);
				V3 colSrcv; colSrcv.SetColor(colSrc);
				Set(u, v, ((colSrcv + colDestv)*0.5f).GetColor());
			}
		}
	}

}


float FrameBuffer::ColorDifference(FrameBuffer *fb2, PPC *ppc2, PPC *ppc1) {

	float err = 0.0f;
	int pixsN = 0;

	M33 M1;
	M1.SetColumn(0, ppc1->a);
	M1.SetColumn(1, ppc1->b);
	M1.SetColumn(2, ppc1->c);
	M33 M2;
	M2.SetColumn(0, ppc2->a);
	M2.SetColumn(1, ppc2->b);
	M2.SetColumn(2, ppc2->c);
	M33 Q = M2.Inverted()*M1;

	for (int v1 = 0; v1 < h; v1++) {
		for (int u1 = 0; u1 < w; u1++) {
			V3 uv1v(.5f + (float)u1, .5f + (float)v1, 1.0f);
			V3 uv2v = Q * uv1v;
			if (uv2v[2] < 0.0f) {
				continue;
			}
			uv2v = uv2v / uv2v[2];
			if (uv2v[0] < 0.0f || uv2v[0] > -0.5f + (float) fb2->w ||
				uv2v[1] < 0.0f || uv2v[1] > -0.5f + (float) fb2->h)
				continue;
			V3 c1; c1.SetColor(Get(u1, v1));
			V3 c2; c2.SetColor(fb2->Get((int)uv2v[0], (int)uv2v[1]));
			err += (c1 - c2)*(c1 - c2);
			pixsN++;
		}
	}

	if (pixsN < 1000)
		return FLT_MAX;

	err /= pixsN;
	return err;

}

void FrameBuffer::SetFromEQRI(PPC *ppc, FrameBuffer *eqrfb) {
	for (int v = 0; v < h; v++) {
		for (int u = 0; u < w; u++) {
			V3 rdir = ppc->GetRay(u, v);
			float alpha = atanf(rdir[2] / rdir[0]);
			float uf; // horizontal coordinate in equirectangular image
			float pif = 3.1415962f;
			if (rdir[0] >= 0.0f) {
				alpha += pif / 2.0f;
				alpha /= pif;
				uf = (float)eqrfb->w / 2.0f + (float)eqrfb->w / 2.0f * alpha;
			}
			else {
				alpha /= pif / 2.0f;
				alpha = (alpha + 1.0f) / 2.0f;
				uf = alpha * (float)eqrfb->w / 2.0f;
			}
			float vf; // vertical coordinate in equirectangular image
			float theta = asinf(rdir[1]);
			vf = (float)eqrfb->h / 2.0f * (1.0f - theta / (pif / 2.0f));
			Set(u, v, scene->panofb0->GetBL(uf, vf));
			eqrfb->Set((int)uf, (int)vf, 0xFF0000FF);
		}
	}
	eqrfb->redraw();
	Fl::check();
	eqrfb->CopyImage(scene->panofb0);
}

unsigned int FrameBuffer::GetBL(float uf, float vf) {

	if (uf < 0.0f || uf >= (float)w || vf < 0.0f || vf >= (float)h)
		return 0xFFFF00FF;

	if (
		uf < 0.5f || uf >= (float)w - 0.5f ||
		vf < 0.5f || vf >= (float)h - 0.5f
		)
		return Get((int)uf, (int)vf);

	int u0 = (int)(uf - 0.5f);
	int v0 = (int)(vf - 0.5f);
	float dx = uf - (.5f + (float)u0);
	float dy = vf - (.5f + (float)v0);

	V3 A; A.SetColor(Get(u0, v0));
	V3 B; B.SetColor(Get(u0+1, v0));
	V3 C; C.SetColor(Get(u0+1, v0+1));
	V3 D; D.SetColor(Get(u0, v0+1));

	V3 retv =
		A*(1.0f - dx)*(1.0f - dy) +
		B*dx*(1.0f - dy) +
		C*dx*dy +
		D*(1.0f - dx)*dy;

	return retv.GetColor();

}

void FrameBuffer::SetFromFisheye(PPC *ppc, FrameBuffer *pfb) {

	for (int v = 0; v < h; v++) {
		for (int u = 0; u < w; u++) {
			V3 rdir = ppc->GetRay(u, v);
			if (rdir[2] > 0.0f) {
				Set(u, v, 0xFF00FF00);
				continue;
			}
//rdir = V3(0.0f, 0.0f, -1.0f);

			float uf, vf; // corresponding point inside panorama
			V3 fvd(0.0f, 0.0f, -1.0f); // fisheye view direction
			float pif = 3.1415962f;
			float alpha = acosf(rdir*fvd)*180.0f / pif;
			float R = (float)pfb->h / 2.0f;
			V3 q(rdir[0], rdir[1], 0.0f);
			q = q.UnitVector();
			// alpha = 0 then r = 0
			// alpha = 90 then r = R
			float r = R * (alpha / 90.0f);
			uf = (float) pfb->w/2.0f + q[0] * r;
			vf = (float) pfb->h/2.0f - q[1] * r;
			Set(u, v, scene->panofb0->GetBL(uf, vf));
			pfb->Set((int)uf, (int)vf, 0xFF0000FF);
		}
	}
	pfb->redraw();
	Fl::check();
	pfb->CopyImage(scene->panofb0);
}
